

/* 
 *  Licensed Materials - Property of IBM
 *  Use Restricted Material of IBM - please refer to the "Samples Code" terms and conditions in the IBM International Program License Agreement (IPLA).
 *  IBM Rational Test RealTime
 *  (c) Copyright IBM Corp. 2003-2007 All Rights Reserved.
 *
 *  US Governement Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */



#include <stdio.h>
#include <string.h>
#include "maitredhote.h"
#include "philosopher.h"
#include "lefthandphilosopher.h"
#include "righthandphilosopher.h"
#include "greedyphilosopher.h"
#include "slowphilosopher.h"
#include "onehandedphilosopher.h"
#include "fork.h"

// ========================================================================
// Returns a new philosopher according to the command line arguments
Philosopher* MaitreDHote::newPhilosopher (int index, int argc, char** argv)
{
  index ++;
  if (index < argc) {
    // Use the user selection
    if (strcmp (argv[index], "left") == 0)
      return new LeftHandPhilosopher(*this);
    else if (strcmp (argv[index], "right") == 0)
      return new RightHandPhilosopher(*this);
    else if (strcmp (argv[index], "greedy") == 0)
      return new GreedyPhilosopher(*this);
    else if (strcmp (argv[index], "slow") == 0)
      return new SlowPhilosopher(*this);
    else if (strcmp (argv[index], "one") == 0)
      return new OneHandedPhilosopher(*this);
    else
      throw new invalidPhilospher;
  }
  else {
    //  Return the default philospher
    return new Philosopher(*this);
  }
}


// ========================================================================
// Constructor
MaitreDHote::MaitreDHote ()
{
}

// ========================================================================
// Creates and seats the philosophers at the table with the forks
void MaitreDHote::seatPhilosophers (int argc, char** argv)
{
#pragma attol att_insert char note[256];
#pragma attol att_insert sprintf (note, "argc = %d ", argc);
#pragma attol att_insert _ATT_USER_NOTE (note)

  // How many philosophers are dining?
  numberOfPhilosophers = argc - 1;

  // create the forks and the philosophers
  theForks = new Fork*[numberOfPhilosophers];
  thePhilosophers = new Philosopher*[numberOfPhilosophers];

  // Allocate the first fork
  theForks[0] = new Fork();

  // then allocate the philosophers and the other forks
  try {
    for (int i = 0; i < numberOfPhilosophers; i++) {
      // Then allocate the philosopher and assign the left fork
      thePhilosophers[i] = newPhilosopher(i, argc, argv);
      thePhilosophers[i]->assignLeftFork (theForks[i]);
      
      // Allocate the right fork if necessary
      if ((i + 1) < numberOfPhilosophers)
	theForks[i + 1] = new Fork();
      
      // Assign the fork to the Philosopher
      thePhilosophers[i]->assignRightFork (theForks[(i + 1) %
						    numberOfPhilosophers]);
    }
  } catch (invalidPhilospher e) {
    fprintf (stderr, "Invalid paramter passed. The parameters must be one of\nleft\nright\n");
    exit (1);
  }
}

// ========================================================================
// Cleanup all allocations and ensure that all mutexes are freed
MaitreDHote::~MaitreDHote()
{
  for (int i = 0; i < numberOfPhilosophers; i++) {
    delete theForks[i];
    delete thePhilosophers[i];
  }
  delete [] theForks;
  delete [] thePhilosophers;
}

// ========================================================================
// Returns true if someone has eaten since the last call to the function
bool MaitreDHote::hasAnyoneEaten()
{
  if (someoneHasEaten) {
    someoneHasEaten = false;
    return (true);
  }
  else
    return false;
}

// ========================================================================
// Returns true if all the diners have finished eating
bool MaitreDHote::allHaveFinished()
{
  bool allFinished = true;
  // All have eaten when there are nolonger any philosophers at the table
  for (int i = 0; i < numberOfPhilosophers; i++)
    allFinished = allFinished && (thePhilosophers[i] == 0);
  return (allFinished);
}

// ========================================================================
// The main thread loop run from the watchdog timer
void MaitreDHote::run ()
{
  int numEaten = 0;
  bool wait = true;
  while (wait) {
    // Sleep for 5 seconds)
    Sleep (5);

    // Has any Philosopher finished
    for (int i = 0; i < numberOfPhilosophers; i++) {
      if (thePhilosophers[i] != 0) {
	if (thePhilosophers[i]->getThread()->isFinished(1)) {
#pragma attol att_insert _ATT_USER_NOTE ("Thread has finished")
	  delete thePhilosophers[i];
	  thePhilosophers[i] = 0;
	}
      }
    }

    // Has anyone eaten
    if (( ! hasAnyoneEaten()) || allHaveFinished()) {
      // Allow 5 seconds for someone to eat.
      if (allHaveFinished() || (numEaten++ == 1000))
	// stop the process
	wait = false;
    }
    else
      numEaten = 0;
  }

  // Status report
  if (allHaveFinished()) {
#pragma attol att_insert _ATT_USER_NOTE ("All have Finished")
    printf ("\nAll philosophers have finished eating\n\n");
  } else {
#pragma attol att_insert _ATT_USER_NOTE ("DEADLOCK")
    printf ("\nDEADLOCK detected, program aborted\n\n");
  }
}

// ========================================================================
// The watchdog thread
DWORD watchDog (LPDWORD lpdwParam)
{
  // Get the pointer to the maitre d'hote instance
  MaitreDHote* pMaitreDHote = (MaitreDHote*)lpdwParam;
  pMaitreDHote->run();
  return (0);
}

// ========================================================================
// Start all the threads and then wait for them all to stop
void MaitreDHote::Dine(int argc, char** argv)
{
  // Create and seat the philosophers at the table with the forks
  seatPhilosophers(argc, argv);

  // Create the watch dog timer
  watchdogThread = CreateThread( 0, 0,
				 (LPTHREAD_START_ROUTINE)watchDog,
				 (LPVOID)this, CREATE_SUSPENDED, 0 );
  if (watchdogThread == (HANDLE)0)
    throw new createWatchdogFailed;

  // Start the watchdog timer
  ResumeThread (watchdogThread);

  // Start the philosophers dining
  for (int i = 0; i < numberOfPhilosophers; i++)
    thePhilosophers[i]->Dine ();


  // Wait for it to terminate
  WaitForSingleObject(watchdogThread, INFINITE);
}

// ========================================================================
// Informs the MaitreDHote that a philosopher is eating
void MaitreDHote::amEating (Philosopher& aPhilosopher)
{
  someoneHasEaten = true;
}
